
-- global hook (everything local be moved inside this)
SaneLFGObj = { bDebug = false };

local	bInitialQuery = true;
local	bPause = false;
local	iRestoreFromParty;
local	CHATFRAME = ChatFrame2;

-- forward declarations
local	CommandHandler;
local	RestoreWindow;

-- override in SaneLFG_Override
local	function	SaneLFMFrameFilter_DataMatches(...)
	local	name, level, zone, class, criteria1, criteria2, criteria3, comment, numPartyMembers, isLFM, classFileName, willBeLeader, willBeTank, willBeHealer, willBeDPS = ...;
	
	if ( LFMFrameColumn1Type == LFMCOLUMN1TYPE_GROUP ) then
		if ( (not numPartyMembers) or (numPartyMembers <= 0) ) then
			return false;
		end
	elseif ( (LFMFrameColumn1Type == LFMCOLUMN1TYPE_INDIVIDUAL)  ) then
		if ( numPartyMembers and (numPartyMembers > 0) ) then
			return false;
		end

		local roleFilterTab = LFMFrame_RoleFilter_Individual;
		if ( not ( (roleFilterTab["leader"] and willBeLeader) or
					(roleFilterTab["tank"] and willBeTank) or
					(roleFilterTab["healer"] and willBeHealer) or
					(roleFilterTab["damage"] and willBeDPS) ) ) then
			return false;
		end
	elseif ( (LFMFrameColumn1Type == LFMCOLUMN1TYPE_ALL)  ) then
		if ( (not numPartyMembers) or (numPartyMembers <= 0) ) then
			local roleFilterTab = LFMFrame_RoleFilter_Individual;
			if ( not ( (roleFilterTab["leader"] and willBeLeader) or
						(roleFilterTab["tank"] and willBeTank) or
						(roleFilterTab["healer"] and willBeHealer) or
						(roleFilterTab["damage"] and willBeDPS) ) ) then
				return false;
			end
		end
	end
	
	return true;
end

--
--
--


local	oInstancesSaved = { Count = 0, Age = 0, Reset = 0, Saved = {} };

local	function	SaneLFGDropdown_BuildSaved(bForce)
	local	Saved = oInstancesSaved.Saved;

	-- check if something is about to run out:
	if (oInstancesSaved.Reset > 0) then
		local	iDelta = time() - oInstancesSaved.Age;
		if (iDelta > oInstancesSaved.Reset) then
			oInstancesSaved = { Count = 0, Age = 0, Reset = 0, Saved = {} };
		end
	end

	local	iMax = GetNumSavedInstances();
	if (bForce or (oInstancesSaved.Count ~= iMax)) then
		local	iReset = 86400;
		Saved = {};

		local	i;
		for i = 1, iMax do
			local	sName, iID, iResetsInSeconds, iDifficulty, bLocked, _, _, bIsRaid, iPlayers = GetSavedInstanceInfo(i);
			if (bLocked) then
				sName = string.gsub(sName, "\194\160", " ");	-- compensate for Ahn'Kahet\190\164 error
				sName = string.gsub(sName, " %-", "-");		-- compensate for Azjol -Nerub error

				sName = strlower(sName);
				if (Saved[sName] == nil) then
					Saved[sName] = {};
				end

				Saved[sName][iPlayers] = iID;

				if (iResetsInSeconds > 0) then
					iReset = math.min(iReset, iResetsInSeconds);
				end
			end
		end

		oInstancesSaved.Age = time();
		oInstancesSaved.Reset = iReset;
		oInstancesSaved.Count = iMax;
		oInstancesSaved.Saved = Saved;
	end

	return Saved;
end

local	function	SaneLFGDropdown_CheckNameInSaved(selectedType, sNameInCombo, oSaved)
	-- only add if we're not locked out already
	local	bSaved = false;
	local	sName = strlower(sNameInCombo);

	-- now, in theory the names *should* be identical, any sane person wouldn't have different names for the same dungeon at different places, right?
	-- never underestimate Blizzard: Ahn'Kahet\194\160:... vs. Ahn'Kahet :..., Azjol-Nerub vs. Azjol -Nerub, Fort pourpe vs. Le fort Pourpre (even different caps!)
	-- insanity, thy name is Blizzard

	sName = string.gsub(sName, "\194\160", " ");	-- compensate for Ahn'Kahet\190\164 error
	if (selectedType == 6) then
		-- easy case: dungeon is displayed as "dungeon (heroic)", but saved as "dungeon"

		-- compensate for region prefixes
		local	iPos = string.find(sName, " %- ");	-- with spaces, so we don't catch "Azjol-Nerub"
		if (iPos) then
			-- Ulduar - Halls of Lightning => Halls of Lightning
			sName = strsub(sName, iPos + 3);
		end

		sName = string.gsub(sName, " %-", "-");		-- compensate for Azjol -Nerub error

		local	iPos = string.find(sName, "%(");
		if (iPos) then
			sName = strsub(sName, 1, iPos - 2);
		end

		if (oSaved[sName]) then
			bSaved = oSaved[sName][5] ~= nil;
		else
			-- it SHOULD be that easy... but Blizzard again breaks stuff for NO reason :-(
			-- names in the LFG can be *dfferent* from those in raid lockout even though they describe the same dungeon
			-- (and naturally we have no access to an ID or something language neutral in both places)
			-- example (locale frFR): Fort pourpre (in LFG) == Le fort Pourpre (in raid lockout)
			-- so: check if lockout name *contains* current name fragment
			local	k, v;
			for k, v in pairs(oSaved) do
				if (string.find(k, sName, 1, true)) then
					bSaved = true;
					break;
				end
			end
		end
	elseif (selectedType == 3) then
		-- slightly more difficult case: must parse string a bit better "dungeon (10 players)"
		local	_, _, sNameSub, sPlayers = string.find(sName, "(.+) %((%d+) .+%)");
		local	iPlayers = tonumber(sPlayers);
--				CHATFRAME:AddMessage("SaneLFG/DEBUG: [" .. index .. "] " .. info.text .. " => " .. tostring(sNameSub) .. "::" .. tostring(iPlayers));
		if (sNameSub and oSaved[sNameSub] and iPlayers and (oSaved[sNameSub][iPlayers] ~= nil)) then
			bSaved = true;
		end
	end

	return bSaved;
end

local	function	SaneLFGDropdown_GetLFGTypeNames(...)
	local info = UIDropDownMenu_CreateInfo();
	local typeDropdown = ...;
	local dropdownID = typeDropdown:GetID();
	local selectedType = UIDropDownMenu_GetSelectedID(typeDropdown);

	local	oSaved = SaneLFGDropdown_BuildSaved();

	local iCnt, index = 1;
	for i = 2, select("#", ...), 2 do
		-- Limit number of displayable names to 20 (40/2)
		if ( iCnt <= 20 ) then
			iCnt = iCnt + 1;

			index = i/2;
			info.text = select(i, ...);
			info.value = select(i+1, ...);
			info.func = LFGNameButton_OnClick;
			info.owner = UIDROPDOWNMENU_OPEN_MENU;
			info.checked = nil;
			info.disabled = nil;
			for j = 1, #LFG_DISABLED_DROPDOWN_NAMES do
				if ( j ~= dropdownID ) then
					if ( LFG_DISABLED_DROPDOWN_NAMES[j].type == selectedType and LFG_DISABLED_DROPDOWN_NAMES[j].name == index ) then
						info.disabled = 1;
					end
				end
			end

			-- show red if we're locked out
			local	bSaved = SaneLFGDropdown_CheckNameInSaved(selectedType, info.text, oSaved);
			if (bSaved) then
				info.text = "|cFFFF6060" .. info.text .. "|r";
			end

			UIDropDownMenu_AddButton(info);
		end
	end
end

local	function	SaneLFGDropdown_GetLFMTypeNames(...)
	local	selectedType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
	local	oSaved = SaneLFGDropdown_BuildSaved();

	local info = UIDropDownMenu_CreateInfo();
	for i=1, select("#", ...), 2 do
		info.text = select(i, ...);
		info.func = LFMNameButton_OnClick;
		info.owner = UIDROPDOWNMENU_OPEN_MENU;
		info.checked = nil;

		-- show red if we're locked out
		local	bSaved = SaneLFGDropdown_CheckNameInSaved(selectedType, info.text, oSaved);
		if (bSaved) then
			info.text = "|cFFFF6060" .. info.text .. "|r";
		end

		UIDropDownMenu_AddButton(info);
	end
end

local	function	SaneLFG_Overrides()
	-- override
	LFMCOLUMN1TYPE_INDIVIDUAL = 1;
	LFMCOLUMN1TYPE_GROUP = 2;
	LFMCOLUMN1TYPE_ALL = 3;

	-- override
	LFMFRAME_DROPDOWN_LIST1 = {
	  { name = INDIVIDUALS,
	      func = function(self)
		UIDropDownMenu_SetSelectedID(LFMFrameDropDown1, self:GetID());
		LFMFrameColumn1Type = LFMCOLUMN1TYPE_INDIVIDUAL;
		LFMFrame.selectedName = nil;
		LFMFrame.selectedLFM = nil;
		LFMFrame_CacheAndUpdate();
	      end,
	  },
	  { name = GROUPS,
	      func = function(self)
		UIDropDownMenu_SetSelectedID(LFMFrameDropDown1, self:GetID());
		LFMFrameColumn1Type = LFMCOLUMN1TYPE_GROUP;
		LFMFrame.selectedName = nil;
		LFMFrame.selectedLFM = nil;
		LFMFrame_CacheAndUpdate();
	      end,
	  },
	  { name = ALL,
	      func = function(self)
		UIDropDownMenu_SetSelectedID(LFMFrameDropDown1, self:GetID());
		LFMFrameColumn1Type = LFMCOLUMN1TYPE_ALL;
		LFMFrame.selectedName = nil;
		LFMFrame.selectedLFM = nil;
		LFMFrame_CacheAndUpdate();
	      end,
	  },
	}

	LFMFrameFilter_DataMatches = SaneLFMFrameFilter_DataMatches;
	Dropdown_GetLFGTypeNames = SaneLFGDropdown_GetLFGTypeNames;
	Dropdown_GetLFMTypeNames = SaneLFGDropdown_GetLFMTypeNames;
end

-- copy (override wouldn't do any good, as it's too late: OnLoad is executed while loading the xml/lua)
local	function SaneLFMFrameDropDown1_OnLoad(self)
  LFMFrameColumn1Type = LFMCOLUMN1TYPE_ALL;
  UIDropDownMenu_Initialize(self, LFMFrameDropDown1_Initialize);
  UIDropDownMenu_SetSelectedID(self, LFMFrameColumn1Type);
  UIDropDownMenu_JustifyText(self, "LEFT")
end

--
--
--

-- inline from XML extracted to functions
local	function	LFMFrameTypeDropDown_OnLoad(self)
	UIDropDownMenu_Initialize(self, LFMFrameTypeDropDown_Initialize);
	UIDropDownMenu_SetSelectedID(self, 1);
	UIDropDownMenu_JustifyText(self, "LEFT");
end

local	function	LFMFrameNameDropDown_OnLoad(self)
	UIDropDownMenu_Initialize(self, LFMFrameNameDropDown_Initialize);
	UIDropDownMenu_JustifyText(self, "LEFT");
end

--
--
--

-- due to taint issues, (might) need to resolve the setting of the menus in a whole different manner
function	SaneLFGObj.MenuEntrySet(bLFG, iPos, iType, iName, bQuery)
	if (bLFG) then
		local	oType = _G["LFGFrameTypeDropDown" .. iPos];
		UIDropDownMenu_Initialize(oType, LFGFrameTypeDropDown_Initialize);
		SetLFGTypeCriteria(oType, iType, true);

		local	oName = _G["LFGFrameNameDropDown" .. iPos];
		UIDropDownMenu_Initialize(oName, _G["LFGFrameNameDropDown" .. iPos .. "_Initialize"]);
		SetLFGNameCriteria(oName, iName, UIDropDownMenu_GetValue(iName), true);
	else
		-- LFM: currently only one choice
		UIDropDownMenu_Initialize(LFMFrameTypeDropDown, LFMFrameTypeDropDown_Initialize);
		SetLFMTypeCriteria(iType);

		UIDropDownMenu_Initialize(LFMFrameNameDropDown, LFMFrameNameDropDown_Initialize);
		UIDropDownMenu_SetSelectedID(LFMFrameNameDropDown, iName);
		SetLookingForMore(iType, iName);
	end
end

--
--
--

local	oPausebutton = CreateFrame("Button", "SaneLFGPauseToggle", LFGFrameRolesBorder, "UIPanelButtonTemplate");
oPausebutton:SetText("Pause");

local	function	SaneLFG_UpdatePauseText()
	if (bPause) then
		oPausebutton:SetText("Resume");
	elseif (iRestoreFromParty) then
		oPausebutton:SetText("Waiting.");
	else
		oPausebutton:SetText("Pause");
	end
end

local	function	SaneLFG_TogglePause()
	if (iRestoreFromParty) then
		iRestoreFromParty = 999;
	else
		CommandHandler("pause");
		SaneLFG_UpdatePauseText();
	end
end

oPausebutton:SetScript("OnClick", SaneLFG_TogglePause);

local	function	SaneLFG_ShowTip(oFrame)
	if (bPause and SaneLFGPerChar.LFG) then
		local	iCnt = #SaneLFGPerChar.LFG;
		if (iCnt > 0) then
			GameTooltip:SetOwner(oFrame, "ANCHOR_BOTTOMRIGHT");
			GameTooltip:AddLine("SaneLFG: Currently PAUSEd.");
			GameTooltip:AddLine(" ");
			GameTooltip:AddLine("Clicking RESUME will (try to) switch to:");
			local	i;
			for i = 1, iCnt do
				local	oLFG = SaneLFGPerChar.LFG[i];
				local	sType = SaneLFGObj.TypeToString(oLFG.Type);
				local	sName = SaneLFGObj.TypeNameToString(oLFG.Type, oLFG.Name);
				GameTooltip:AddLine("[LFG " .. i .. "] " .. sType .. ": " .. sName);
			end
			if (SaneLFGPerChar.LFM) then
				local	oLFM = SaneLFGPerChar.LFM;
				local	sType = SaneLFGObj.TypeToString(oLFM.Type);
				local	sName = SaneLFGObj.TypeNameToString(oLFM.Type, oLFM.Name);
				GameTooltip:AddLine(" ");
				GameTooltip:AddLine("[LFM] " .. sType .. ": " .. sName);
			end
			GameTooltip:Show();
		end
	end
end

oPausebutton:SetScript("OnEnter", SaneLFG_ShowTip);
oPausebutton:SetScript("OnLeave", GameTooltip_Hide);

oPausebutton:SetPoint("TOPRIGHT", LFGFrameRolesBorder, "TOPRIGHT", -8, -8);
oPausebutton:SetWidth(80);
oPausebutton:SetHeight(25);
oPausebutton:Show();

--
--
--

SaneLFGGlobal = { Cfg = {} };
SaneLFGPerChar = { Cfg = {}, LFG = {} };

-- states need symbolic flags
SaneLFGObj.States =
	{
		PLAYER_IN_WORLD = 1,
		LOGON_COMPLETED = 2,
		LFG_INFO_COMPLETE = 4,
	};

local	iState = 0;
local	oLFGData;

function	SaneLFGObj.TypeToString(iType)
	local	sType = "?" .. iType;
	if (oLFGData and oLFGData.Types and oLFGData.Types[iType]) then
		sType = oLFGData.Types[iType];
	end

	return sType;
end

function	SaneLFGObj.TypeNameToString(iType, iName)
	if (iName == 0) then
		return " -- none -- ", true;
	end

	local	sName = "?" .. iName;
	local	bValid = false;
	if (oLFGData and oLFGData.Names and oLFGData.Names[iType]) then
		sName = oLFGData.Names[iType][iName * 2 - 1] or sName;
		bValid = true;
	end

	return sName, bValid;
end

function	SaneLFGObj.TypeNameCount(iType)
	local	iNum = 0;
	if (oLFGData and oLFGData.Names and oLFGData.Names[iType]) then
		iNum = #oLFGData.Names[iType] / 2;
	end

	return iNum;
end

function	SaneLFGObj.TypeNameValid(iType, iName)
	local	bValid = false;
	if (oLFGData and oLFGData.Names and oLFGData.Names[iType]) then
		bValid = nil ~= oLFGData.Names[iType][iName * 2 - 1];
	end

	return bValid;
end

local	function	CheckLFGInitComplete(sEvent)
	if (bit.band(iState, 4) ~= 4) then
		oLFGData = {};
		oLFGData.Types = { GetLFGTypes() };
		oLFGData.Names = {};
	end

	if (oLFGData and (next(oLFGData.Types) ~= nil)) then
		local	oFirst = {};

		local	bComplete = true;
		local	i;
		for i = 2, #oLFGData.Types do
			if (oLFGData.Names[i] and (next(oLFGData.Names[i]) ~= nil)) then
				oFirst[i] = true;
			else
				oLFGData.Names[i] = { GetLFGTypeEntries(i) };
				if (next(oLFGData.Names[i]) ~= nil) then
					oFirst[i] = true;
				else
					oFirst[i] = false;
				end
			end
		end

		if (oLFGData.CompletedWith == nil) then
			local	bComplete = true;
			for i = 1, 4 do
				local	oData;
				if (i < 4) then
					oData = SaneLFGPerChar.LFG[i];
				else
					oData = SaneLFGPerChar.LFM;
				end

				if (oData and oData.Type and oData.Name and (oData.Name ~= 0)) then
					if (not oFirst[oData.Type]) then
						if ((oData.Type == 6) and (UnitLevel("player") >= 73) and (UnitLevel("player") <= 79)) then
							-- if BC heroic dungeon list gets empty (73-79), we lose the category in one single step
							-- so, don't count this as missing in the range
						else
							bComplete = false;
							break
						end
					end
				end
			end

			if (bComplete) then
				oLFGData.CompletedWith = sEvent;
				iState = bit.bor(iState, 4);
			end
		end
	end
end

local	function	UpdateChoiceToAvailability()
	local	function	MapData(sPrefix, oData)
			if (oData and oData.Name and (oData.Name ~= 0)) then
				local	iType = oData.Type;
				local	iOld = oData.Name;
				oData.Name = 0;

				local	oEntries = oLFGData.Names[iType];
				if (oEntries and (next(oEntries) ~= nil)) then
					local	iIndex, sName;
					for iIndex, sName in pairs(oEntries) do
						-- regions have a *blank* English text... => can't go for neutral desc :(
						if ((iIndex % 2 == 1) and (sName == oData.DescLoc)) then
							oData.Name = (iIndex + 1) / 2;

							local	sMsg = "Mapped '" .. oData.DescLoc .. "' from " .. iOld .. " to " .. oData.Name .. ".";
							CHATFRAME:AddMessage("SaneLFG: " .. sPrefix .. " - " .. sMsg);
							break;
						end
					end
				end

				if (oData.Name == 0) then
					CHATFRAME:AddMessage("SaneLFG: Couldn't find '" .. oData.DescLoc .. "' again. Dropping it...");
				end
			end
		end

	local	i, oData;
	if (SaneLFGPerChar.LFG) then
		local	bCompress = false;
		for i = 1, 3 do
			oData = SaneLFGPerChar.LFG[i];
			if (oData) then
				MapData(" LFG #" .. i, oData);
				if (oData.Name == 0) then
					bCompress = true;
				end
			end
		end
		if (bCompress) then
			local	oLFGNew, iNew = {}, 1;
			for i = 1, 3 do
				oData = SaneLFGPerChar.LFG[i];
				if (oData and (oData.Name ~= 0)) then
					oLFGNew[iNew] = oData;
					iNew = iNew + 1;
				end
				SaneLFGPerChar.LFG[i] = nil;
			end
			SaneLFGPerChar.LFG = oLFGNew;
		end
	end

	oData = SaneLFGPerChar.LFM;
	MapData(" LFM", oData);

	if (SaneLFGPerChar.LFMSet) then
		local	oRemove = {};
		local	iCnt = #SaneLFGPerChar.LFMSet;
		for i = 1, iCnt do
			oData = SaneLFGPerChar.LFMSet[i];
			if (oData) then
				MapData(" LFG set entry #" .. i, oData);
				if (oData.Name == 0) then
					tinsert(oRemove, i);
				end
			end
		end

		if (#oRemove > 0) then
			for i = #oRemove, 1, -1 do
				tremove(SaneLFGPerChar.LFMSet, oRemove[i]);
			end
		end
	end
end

local	bInit = true;
local	bTypesCompleteFirst = true;
local	sVersionNumber, sVersionDate;

local	bStore = false;
local	bStoreBeforePartyChange;
local	bDing = false;
local	bInParty;

local	iElapsedSecond = 0;
local	iPartyChanged;
local	bUpdate = false;

local	function SaneLFGEvent(oFrame, sEvent, ...)
	if (sEvent == "ADDON_LOADED") then
		local	sWhich = ...;
		if (sWhich == "SaneLFG") then
			SaneLFGPerChar.LFGHistory = nil;
			SaneLFGPerChar.LFMHistory = nil;

			sVersionNumber = GetAddOnMetadata("SaneLFG", "Version");
			if (sVersionNumber == nil) then
				sVersionNumber = "???";
			end
			sVersionDate = GetAddOnMetadata("SaneLFG", "X-Date");
			if (sVersionDate == nil) then
				sVersionDate = "???";
			end

			if (SaneLFGGlobal.Cfg.PauseSave) then
				bPause = SaneLFGPerChar.Cfg.PauseStatus or false;
			end

			if (SaneLFGPerChar.Cfg.PauseOnLogin) then
				bPause = true;
			end

			SaneLFGObj.bDebug = SaneLFGGlobal.Cfg.Debug == true;

			SLASH_SANELFG1 = "/sanelfg";
			SlashCmdList["SANELFG"] = CommandHandler;
		end

		return
	end

	CheckLFGInitComplete(Event);

	if (SaneLFGObj.bDebug) then
		CHATFRAME:AddMessage("SaneLFG/DEBUG: event = " .. sEvent);
		if (sEvent == "UPDATE_LFG_TYPES") then
			-- are we told which?
			local	sOut, k, v = "";
			for k, v in pairs( { ... } ) do
				sOut = sOut .. ", -" .. tostring(k) .. "- = [" .. tostring(v) .. "]";
			end
			if (sOut ~= "") then
				CHATFRAME:AddMessage("SaneLFG/DEBUG: args = " .. strsub(sOut, 3));
			else
				CHATFRAME:AddMessage("SaneLFG/DEBUG: no args.");
			end
		end
	end

	if (sEvent == "PLAYER_ENTERING_WORLD") then
		iState = bit.bor(iState, 1);
	end
	if ((sEvent == "LFG_UPDATE") or (sEvent == "PLAYER_LOGIN")) then
		iState = bit.bor(iState, 2);
	end

	if (bInit and (bit.band(iState, 3) == 3)) then
		bInit = false;
		CHATFRAME:AddMessage("SaneLFG " .. sVersionNumber .. " (" .. sVersionDate .. ") active. Use /sanelfg to issue commands.");
		if (SaneLFGGlobal.Cfg.PauseSave and SaneLFGPerChar.Cfg.PauseStatus) then
			SaneLFG_UpdatePauseText();
			CHATFRAME:AddMessage("SaneLFG: Pause status is saved across sessions and was active. Not restoring anything!");
		end

		SaneLFG_Overrides();
	end

	if (bTypesCompleteFirst and (bit.band(iState, 4) == 4)) then
		bTypesCompleteFirst = false;
		if (SaneLFGPerChar.LFMSet and (#SaneLFGPerChar.LFMSet > 0)) then
			local	bFixed, bMissing = false, false;
			local	iCnt, i = #SaneLFGPerChar.LFMSet;
			for i = 1, iCnt do
				local	oData = SaneLFGPerChar.LFMSet[i];
				if (oData.DescLoc == nil) then
					local	sName, bValid = SaneLFGObj.TypeNameToString(oData.Type, oData.Name);
					if (bValid) then
						bFixed = true;
						oData.DescLoc = sName;
					else
						bMissing = true;
					end
				end
			end

			if (bFixed) then
				CHATFRAME:AddMessage("SaneLFG: Fixed missing data in LFM set.");
			end
			if (bMissing) then
				CHATFRAME:AddMessage("SaneLFG: Couldn't find some missing data for LFM set.");
			end
		end
	end

	if (sEvent == "UPDATE_LFG_LIST") then
		-- this is only sent when the LFGQuery criteria changed and the new results are available
		-- this is NOT sent when modified results for the same criteria are available
		bInitialQuery = false;
		if (SaneLFGObj.LFMFrame_FullCacheAndUpdate) then
			SaneLFGObj.LFMFrame_FullCacheAndUpdate(true);
		end
	end

	if (sEvent == "PLAYER_LEVEL_UP") then
		iState = bit.band(iState, 3);
		bStore = false;
		bDing = true;
		CHATFRAME:AddMessage("SaneLFG: DING!");
	end

	if (iState == 255) then
		if ((sEvent == "PARTY_MEMBERS_CHANGED") or (sEvent == "RAID_ROSTER_UPDATE")) then
			-- due to the weird logic of PARTY_MEMBERS_CHANGED, we have to delay a response until we can be pretty sure that all events are done
			iPartyChanged = 0;
			bUpdate = true;
			if (bStoreBeforePartyChange == nil) then
				bStoreBeforePartyChange = bStore;
			end
			bStore = false;
		elseif (sEvent == "PARTY_MEMBERS_CHANGED_RESPONSE") then
			if (bStoreBeforePartyChange ~= nil) then
				bStore = bStoreBeforePartyChange;
				bStoreBeforePartyChange = nil;
			end

			if (bInParty and (GetNumPartyMembers() == 0) and (GetNumRaidMembers() == 0)) then
				iState = 7;
				bStore = false;

				if (SaneLFGGlobal.Cfg.PauseOnLeavingGroups) then
					if (not bPause) then
						CHATFRAME:AddMessage("SaneLFG: You left a group. Clearing all LFG/LFM modes, going into PAUSE state. Click on RESUME to restore.");
					end

					bPause = true;
					SaneLFG_UpdatePauseText();
				elseif (not bPause) then
					bUpdate = true;
					iRestoreFromParty = 0;
					SaneLFG_UpdatePauseText();
					CHATFRAME:AddMessage("SaneLFG: You left a group. Will (try to) restore LFG/LFM modes in 30 seconds...");
				else
					CHATFRAME:AddMessage("SaneLFG: You left a group, keeping PAUSEd state from before it.");
				end

				LFGParentFrameTab1_OnClick();
				LFGFrameClearAllButton_OnClick();
				ClearLookingForMore();

				-- need to resend LFGQuery, which can only be done in response to a hardware event... :-(
				bInitialQuery = true;

				-- update locked instances
				SaneLFGDropdown_BuildSaved(true);
			end

			bInParty = (GetNumPartyMembers() > 0) or (GetNumRaidMembers() > 0);
		end
	end

	if (iState == 7) then
		-- CHATFRAME:AddMessage("SaneLFG/DEBUG: first = " .. oLFGData.CompletedWith);

		if (bDing) then
			bDing = false;
			CHATFRAME:AddMessage("SaneLFG: Congratulations, you gained a level! Trying to match the previous LFG/LFM choices into new selection...");
			UpdateChoiceToAvailability();
		end

		-- moved from RestoreWindow: need this to be properly initialized in all cases
		if (type(LFMFrameColumn1Type) == "boolean") then
			CHATFRAME:AddMessage("SaneLFG: Selecting '" .. ALL .. "' as default.");
			SaneLFMFrameDropDown1_OnLoad(LFMFrameDropDown1);
		end

		if (bPause) then
			iState = 255;
		elseif (iRestoreFromParty == nil) then
			RestoreWindow(sEvent);
		end
	end

	if (sEvent == "PLAYER_LEAVING_WORLD") then
		iState = bit.band(iState, 6);
		bStore = false;
	end
end

local	function	SaneLFGUpdate(oFrame, iElapsed)
	if (not bUpdate) then
		return
	end

	-- only check every half a second
	iElapsedSecond = (iElapsedSecond or 0) + iElapsed;
	if (iElapsedSecond < 0.5) then
		return
	end

	bUpdate = false;

	-- stupid PARTY_MEMBERS_CHANGED problem...
	if (iPartyChanged) then
		iPartyChanged = iPartyChanged + iElapsedSecond;
		if (iPartyChanged > 5) then
			iPartyChanged = nil;
			SaneLFGEvent(oFrame, "PARTY_MEMBERS_CHANGED_RESPONSE");
		else
			bUpdate = true;
		end
	end

	-- delay restore, people are so easily confused :-(
	if (iRestoreFromParty) then
		iRestoreFromParty = iRestoreFromParty + iElapsedSecond;
		if (iRestoreFromParty > 30) then
			iRestoreFromParty = nil;
			SaneLFG_UpdatePauseText();
			SaneLFGEvent(oFrame, "PARTY_LEFT_RESTORE");
		else
			bUpdate = true;
		end
	end

	iElapsedSecond = 0;
end

local	oEventframe = CreateFrame("Frame", nil, UIParent);

oEventframe:RegisterEvent("ADDON_LOADED");

oEventframe:RegisterEvent("PLAYER_LOGIN");
oEventframe:RegisterEvent("PLAYER_ENTERING_WORLD");
oEventframe:RegisterEvent("PLAYER_LEAVING_WORLD");

-- oEventframe:RegisterEvent("UNIT_PORTRAIT_UPDATE");
-- oEventframe:RegisterEvent("UNIT_DYNAMNIC_FLAGS");
-- oEventframe:RegisterEvent("UNIT_FLAGS");

oEventframe:RegisterEvent("PLAYER_LEVEL_UP");
oEventframe:RegisterEvent("RAID_ROSTER_UPDATE");
oEventframe:RegisterEvent("PARTY_MEMBERS_CHANGED");
oEventframe:RegisterEvent("PARTY_LEADER_CHANGED");

oEventframe:RegisterEvent("MEETINGSTONE_CHANGED");
oEventframe:RegisterEvent("UPDATE_LFG_TYPES");
oEventframe:RegisterEvent("UPDATE_LFG_LIST");
oEventframe:RegisterEvent("LFG_UPDATE");

oEventframe:SetScript("OnEvent", SaneLFGEvent);
oEventframe:SetScript("OnUpdate", SaneLFGUpdate);

local	function	SaneLFGFrame_Update()
	local	bChange = false;
	if (bStore and not bPause and (GetNumPartyMembers() == 0) and (GetNumRaidMembers() == 0) and LFGFrame:IsVisible()) then
		-- CHATFRAME:AddMessage("SaneLFG: LFGFrame_Update()... (" .. tostring(bStore) .. ")");
		local	t, n = {}, {};
		t[1], n[1], t[2], n[2], t[3], n[3] = GetLookingForGroup();

		local	iFrom, i = 1;
		for i = 1, 3 do
			if (n[i] ~= 0) then
				local	sType = SaneLFGObj.TypeToString(t[i]);
				local	sName = SaneLFGObj.TypeNameToString(t[i], n[i]);
				if ((SaneLFGPerChar.LFG[i] == nil) or (n[i] ~= SaneLFGPerChar.LFG[i].Name)) then
					bChange = true;
					CHATFRAME:AddMessage("SaneLFG: Storing new choice #" .. i .. " on LFG tab... (" .. sType .. ": " .. sName .. ")");
				end

				SaneLFGPerChar.LFG[i] = { Type = t[i], Name = n[i], DescLoc = sName };
				iFrom = i + 1;
			else
				break;
			end
		end

		for i = iFrom, 3 do
			if (SaneLFGPerChar.LFG[i]) then
				bChange = true;
				CHATFRAME:AddMessage("SaneLFG: Dropping old choice #" .. i .. " on LFG tab... (" .. (SaneLFGPerChar.LFG[i].Type or "--") .. "/" .. (SaneLFGPerChar.LFG[i].Name or "--") .. " <" .. (SaneLFGPerChar.LFG[i].DescLoc or "???") .. ">)");
			end

			SaneLFGPerChar.LFG[i] = nil;
		end
	end

	if (SaneLFGObj.bDebug and bChange) then
		if (SaneLFGPerChar.LFGHistory == nil) then
			SaneLFGPerChar.LFGHistory = {};
		end

		local	iPos = #SaneLFGPerChar.LFGHistory + 1;
		SaneLFGPerChar.LFGHistory[iPos] = { Callstack = debugstack(), LFG = {} };
		local	oLFGHistory, i = SaneLFGPerChar.LFGHistory[iPos].LFG;
		for i = 1, 3 do
			local	oLFG = SaneLFGPerChar.LFG[i];
			if (oLFG) then
				oLFGHistory[i] = {};

				local	k, v;
				for k, v in pairs(oLFG) do
					oLFGHistory[i][k] = v;
				end
			end
		end
	end
end

hooksecurefunc("LFGFrame_Update", SaneLFGFrame_Update);

--
--
--

local	bLFMFirstShow = true;

local	function	LFMButton_GetComment(oButton, bNoCreate)
	local	sKey = "Comment";

	if (not bNoCreate and (oButton[sKey] == nil)) then
		local	oParent = _G[oButton:GetName() .. "Level"];
		if (oParent) then
			local	oFrame = oButton:CreateFontString(oButton:GetName() .. sKey, "BORDER", "GameFontHighlightSmall");
			oFrame:SetPoint("TOPRIGHT", oParent, "TOPLEFT");
			oFrame:SetWidth(10);
			oFrame:SetHeight(14);

			local	oColor = HIGHLIGHT_FONT_COLOR;
			if (SaneLFGGlobal.Cfg.CommentColor) then
				oColor = SaneLFGGlobal.Cfg.CommentColor;
			end
			oFrame:SetTextColor(oColor.r, oColor.g, oColor.b);
			oFrame:SetJustifyH("RIGHT");

			-- CHATFRAME:AddMessage("SaneLFG/DEBUG: Created comment marker " .. oFrame:GetName() .. ".");
			oButton[sKey] = oFrame;
		end
	end

	return oButton[sKey];
end

local	function	LFMButton_GetCount(oButton, sRole, bNoCreate)
	local	sKey = sRole .. "Cnt";

	if (not bNoCreate and (oButton[sKey] == nil)) then
		local	oFrame = oButton:CreateFontString(oButton:GetName() .. sKey, "BORDER", "GameFontHighlightSmall");
		oFrame:SetPoint("TOPLEFT", oButton[sRole .. "Icon"]);
		oFrame:SetPoint("BOTTOMRIGHT", oButton[sRole .. "Icon"]);
		oButton[sKey] = oFrame;
	end

	return oButton[sKey];
end

local	fOriginalCallback;
local	fReplacedCallback;
local	sCalledViaLFMFrameUpdate = "[C]: in function `LFMFrame_Update'";
local	iOffsetValid = 0;

local	function	SaneLFMFrame_Update()
	local	iType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);

	local	bFullMixed = false;
	if ((iType == 1) and
	    (LFMFrameColumn1Type == LFMCOLUMN1TYPE_ALL)) then
		-- show everything we know
		if (SaneLFGObj.LFMFrame_FullUpdate) then
			if (fOriginalCallback == nil) then
				fOriginalCallback = LFMListScrollFrame:GetScript("OnVerticalScroll");
				if (SaneLFGObj.bDebug) then
					CHATFRAME:AddMessage("SaneLFG/DEBUG: Storing original OnVerticalScroll.");
				end
			end
			if (fReplacedCallback == nil) then
				fReplacedCallback = function(self, offset)
						local	sFromUpdate = debugstack(7, 1, 0);
						if (strsub(sFromUpdate, 1, strlen(sCalledViaLFMFrameUpdate)) == sCalledViaLFMFrameUpdate) then
							if (SaneLFGObj.bDebug) then
								ChatFrame3:AddMessage("Skipping scroll offset reset.");
							end
							if (iOffsetValid) then
								FauxScrollFrame_OnVerticalScroll(self, iOffsetValid, LFM_BUTTONHEIGHT, SaneLFMFrame_Update);
							end
						else
							iOffsetValid = offset;
							FauxScrollFrame_OnVerticalScroll(self, offset, LFM_BUTTONHEIGHT, SaneLFMFrame_Update);
						end
					end
				if (SaneLFGObj.bDebug) then
					CHATFRAME:AddMessage("SaneLFG/DEBUG: Defining replacement OnVerticalScroll.");
				end
			end

			if (fReplacedCallback ~= LFMListScrollFrame:GetScript("OnVerticalScroll")) then
				CHATFRAME:AddMessage("SaneLFG/DEBUG: Switching to replacement OnVerticalScroll.");
				LFMListScrollFrame:SetScript("OnVerticalScroll", fReplacedCallback);
			end

			bFullMixed = true;
			SaneLFGObj.LFMFrame_FullUpdate();
		end
	end

	if (not bFullMixed) then
		if ((fOriginalCallback ~= nil) and (fOriginalCallback ~= LFMListScrollFrame:GetScript("OnVerticalScroll"))) then
			CHATFRAME:AddMessage("SaneLFG/DEBUG: Resetting to original OnVerticalScroll.");
			LFMListScrollFrame:SetScript("OnVerticalScroll", fOriginalCallback);
		end
	end

	local	i;
	for i = 1, LFGS_TO_DISPLAY, 1 do
		local	button = _G["LFMFrameButton" .. i];
		if (not bFullMixed) then
			button.isTitle = nil;
			button.oIndicesGroup = nil;
			button.iEntry = nil;
		end

		if (button:IsShown()) then
			do
				local	oCnt = LFMButton_GetCount(button, "Damage", true);
				if (oCnt) then
					oCnt:SetText("");
				end
				local	oCnt = LFMButton_GetCount(button, "Tank", true);
				if (oCnt) then
					oCnt:SetText("");
				end
				local	oCnt = LFMButton_GetCount(button, "Healer", true);
				if (oCnt) then
					oCnt:SetText("");
				end

				if (button.comment and (button.comment ~= "")) then
					local	oComment = LFMButton_GetComment(button);
					oComment:SetText("+");
				else
					local	oComment = LFMButton_GetComment(button, true);
					if (oComment) then
						oComment:SetText("");
					end
				end
			end

			if (button.MembersFontString:IsShown()) then
				if ((LFMFrameColumn1Type == LFMCOLUMN1TYPE_ALL) and not button.isTitle) then
					button.MembersFontString:SetPoint("LEFT", button.LeaderIcon, "LEFT");			-- at leader icon
				else
					button.MembersFontString:SetPoint("LEFT", "LFMFrameButton" .. i .. "Class", "RIGHT", 10, 0);	-- regular place
				end

				local	comment = button.comment;
				if (comment and (comment ~= "")) then
					local	iDPS, iTank, iHeal = 0, 0, 0;
					if (strsub(comment, 1, 1) == "?") then
						local	sDPS, sTank, sHeal = string.match(comment, "%? *D *(%d) *T *(%d) *H *(%d) *%-");
						if (sDPS and sHeal and sTank) then
							iDPS = tonumber(sDPS);
							iTank = tonumber(sTank);
							iHeal = tonumber(sHeal);
						end
					else
						comment = strlower(comment);

						local	oLocales = SaneLFGLocales.Current;
						if (oLocales == nil) then
							oLocales = SaneLFGLocales["enUS"];
							SaneLFGLocales.Current = oLocales;
						end

						local	oLocale, i, sTest = oLocales.DPS;
						for i, sTest in pairs(oLocale) do
							if (string.match(comment, sTest)) then
								iDPS = - 0.5;
								sDPS = string.match(comment, "(%d+) *" .. sTest);
								if (sDPS) then
									iDPS = - tonumber(sDPS);
								end
							end
						end

						oLocale = oLocales.Tank;
						for i, sTest in pairs(oLocale) do
							if (string.match(comment, sTest)) then
								iTank = - 0.5;
								sTank = string.match(comment, "(%d+) *" .. sTest);
								if (sTank) then
									iTank = - tonumber(sTank);
								end
							end
						end

						oLocale = oLocales.Heal;
						for i, sTest in pairs(oLocale) do
							if (string.match(comment, sTest)) then
								iHeal = - 0.5;
								sHeal = string.match(comment, "(%d+) *" .. sTest);
								if (sHeal) then
									iHeal = - tonumber(sHeal);
								end
							end
						end
					end

					if (math.abs(iDPS) + math.abs(iTank) + math.abs(iHeal) > 0) then
						-- if we're looking at a 5-man group, we should even be able to tell how many...
						if ((iType == 2) or (iType == 6)) then
							local	iMissing = 4;
							if (button.partyMembers and (button.partyMembers > 0)) then
								iMissing = iMissing - button.partyMembers;

								if (iTank ~= 0) then
									if (iTank == - 0.5) then
										iMissing = iMissing - 1;
										iTank = 1;
									else
										iMissing = iMissing - iTank;
									end
								end

								if (iHeal ~= 0) then
									if (iHeal == - 0.5) then
										iMissing = iMissing - 1;
										iHeal = 1;
									else
										iMissing = iMissing - iHeal;
									end
								end

								if ((iDPS == - 0.5) and (iMissing > 0)) then
									iDPS = iMissing;
								end
							end
						end

						local	oCnt = LFMButton_GetCount(button, "Damage");
						if (oCnt) then
							if (iDPS > 0) then
								oCnt:SetText(iDPS);
							elseif (iDPS < 0) then
								if (iDPS == - 0.5) then
									oCnt:SetText("?");
								else
									oCnt:SetText((- iDPS) .. "+");
								end
							else
								oCnt:SetText("");
							end
						end
						local	oCnt = LFMButton_GetCount(button, "Tank");
						if (oCnt) then
							if (iTank > 0) then
								oCnt:SetText(iTank);
							elseif (iTank < 0) then
								if (iTank == - 0.5) then
									oCnt:SetText("?");
								else
									oCnt:SetText((- iTank) .. "+");
								end
							else
								oCnt:SetText("");
							end
						end
						local	oCnt = LFMButton_GetCount(button, "Healer");
						if (oCnt) then
							if (iHeal > 0) then
								oCnt:SetText(iHeal);
							elseif (iHeal < 0) then
								if (iHeal == - 0.5) then
									oCnt:SetText("?");
								else
									oCnt:SetText((- iHeal) .. "+");
								end
							else
								oCnt:SetText("");
							end
						end
					end
				end
			end
		end
	end

	local	bChange = false;
	local	bDo = not bLFMFirstShow and bStore and not bPause and LFMFrame:IsVisible();
	local	bGotLFG = (SaneLFGPerChar.LFG ~= nil) and (#SaneLFGPerChar.LFG > 0);
	if (bGotLFG) then
		local	_, n1, _, n2, _, n3 = GetLookingForGroup();
		if (n1 + n2 + n3 == 0) then
			bGotLFG = false;
			bDo = false;
		end
	else
		bDo = false;
	end

	if (bDo) then
		-- CHATFRAME:AddMessage("SaneLFG: LFMFrame_Update()... (" .. tostring(bStore) .. ")");

		--local type1, name1, type2, name2, type3, name3,
		--	lfmType, lfmName, comment, queued, lfgStatus, lfmStatus = GetLookingForGroup();
		-- GetLookingForGroup() returns lfmName *only* if in a group (yeah, that totally makes sense... NOT)
		-- read the dialog value directly

		local selectedLFMType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
		local selectedLFMName = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
		if (selectedLFMName and (selectedLFMName ~= 0)) then
			local	sType = SaneLFGObj.TypeToString(selectedLFMType);
			local	sName = SaneLFGObj.TypeNameToString(selectedLFMType, selectedLFMName);

			if ((SaneLFGPerChar.LFM == nil) or (selectedLFMName ~= SaneLFGPerChar.LFM.Name)) then
				bChange = true;
				CHATFRAME:AddMessage("SaneLFG: Storing new choice on LFM tab... (" .. sType .. ": " .. sName .. ")");
			end

			local	oEntries = { GetLFGTypeEntries(selectedLFMType) };

			SaneLFGPerChar.LFM = { Type = selectedLFMType, Name = selectedLFMName, DescLoc = sName };
		elseif (SaneLFGPerChar.LFM) then
			bChange = true;
			CHATFRAME:AddMessage("SaneLFG: Resetting to no choice on LFM tab.");
			SaneLFGPerChar.LFM = nil;
		end
	end

	if (bLFMFirstShow and (iState >= 7) and LFMFrame:IsVisible()) then
		bLFMFirstShow = false;
		if (bGotLFG and SaneLFGPerChar.LFM) then
			SaneLFGObj.MenuEntrySet(false, nil, SaneLFGPerChar.LFM.Type, SaneLFGPerChar.LFM.Name);
		end
	end

	if (SaneLFGObj.bDebug) then
		if (bChange) then
			if (SaneLFGPerChar.LFMHistory == nil) then
				SaneLFGPerChar.LFMHistory = {};
			end

			local	iPos = #SaneLFGPerChar.LFMHistory + 1;
			SaneLFGPerChar.LFMHistory[iPos] = { Callstack = debugstack() };
			if (SaneLFGPerChar.LFM) then
				local	k, v;
				for k, v in pairs(SaneLFGPerChar.LFM) do
					SaneLFGPerChar.LFMHistory[iPos][k] = v;
				end
			end
		end
	end
end

hooksecurefunc("LFMFrame_Update", SaneLFMFrame_Update);

--
--
--

local	function	SaneSetLFGComment(sWhich)
	if (bStore and not bPause and LFMFrame:IsVisible()) then
		if ((GetNumPartyMembers() == 0) and (GetNumRaidMembers() == 0)) then
			SaneLFGPerChar.Comment = sWhich;
			if (sWhich == "") then
				SaneLFGPerChar.Comment = nil;
			end
		end
	end
end

hooksecurefunc("SetLFGComment", SaneSetLFGComment);

--
--
--

local	function	SaneToggleLFGParentFrame(oFrame)
--	CHATFRAME:AddMessage("SaneLFG/DEBUG: ToggleLFGParentFrame()...");
	if ((iState == 255) and LFGParentFrame:IsVisible()) then
		if (bInitialQuery) then
			local	Type = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
			local	Name = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
			if (Name and (Name ~= 0)) then
				if (CanSendLFGQuery(Type, Name)) then
					bInitialQuery = false;
					LFGQuery(Type, Name);
				else
					CHATFRAME:AddMessage("SaneLFG: Cannot send LFM update request yet, have to wait...");
				end
			end
		end

		if (SaneLFGObj.LFMFrame_FullCacheAndUpdate) then
			SaneLFGObj.LFMFrame_FullCacheAndUpdate();
		end
	end
end

hooksecurefunc("ToggleLFGParentFrame", SaneToggleLFGParentFrame);

--
--
--

function	CommandHandler(sLine)
	if (sLine == "reset") then
		CHATFRAME:AddMessage("SaneLFG: Resetting stored LFG/LFM settings... (comment will be kept)");
		SaneLFGPerChar.LFG = {};
		SaneLFGPerChar.LFM = nil;
	elseif (sLine == "pause") then
		if (iState < 7) then
			CHATFRAME:AddMessage("SaneLFG: Cannot pause/unpause, not in a stable state yet.");
		else
			bPause = not bPause;
			if (SaneLFGGlobal.Cfg.PauseSave) then
				SaneLFGPerChar.Cfg.PauseStatus = bPause;
			end

			if (bPause) then
				CHATFRAME:AddMessage("SaneLFG: Pausing LFG and resetting selected instances...");
				LFGFrameClearAllButton_OnClick();
				ClearLookingForMore();
				SetLFMType(1);
			else
				CHATFRAME:AddMessage("SaneLFG: Unpausing LFG and restoring previously selected instances...");
				RestoreWindow("CommandHandler");
			end
		end
	elseif (sLine == "lfxstatus") then
		SaneLFGObj.LFGStatus();
	elseif (sLine == "config pausesave") then
		SaneLFGGlobal.Cfg.PauseSave = not SaneLFGGlobal.Cfg.PauseSave;
		if (SaneLFGGlobal.Cfg.PauseSave) then
			CHATFRAME:AddMessage("SaneLFG: Pause/Restore status is now saved.");
		else
			CHATFRAME:AddMessage("SaneLFG: Pause/Restore status is now reset on relogging.");
		end
	elseif (sLine == "validate") then
		UpdateChoiceToAvailability();
	else
		CHATFRAME:AddMessage("SaneLFG: Unknown command '" .. sLine .. "'.\n'/sanelfg pause' toggles pause-status.\n'/sanelfg reset' resets the stored LFG/LFM choices to none.\n'/sanelfg config pausesave' toggles if pause status is stored across logins.\n'/sanelfg lfxstatus' displays the current lfg/lfm status as seen by the client.\n'/sanelfg validate' re-validates your choices against the available options.");
	end
end

function	RestoreWindow(sEvent)
	if (iState < 7) then
		return
	end

	local	iStateNew = 255;
	local	bStoreOld = bStore;
	local	bStoreNew = true;

	bStore = false;

	if (SaneLFGObj.bDebug) then
		-- toggling unavoidable taint, but as we can't even add "all" without tainting anyways...
		CHATFRAME:AddMessage("SaneLFG: |cFFFF0000Taint issue still pending...|r Add at least a message with workaround!");
	end

	bLFMFirstShow = false;
	ToggleLFGParentFrame();
	ToggleLFGParentFrame();
	bLFMFirstShow = true;

	-- LFG parts

	local	bGotLFG = false;
	local	sComment;

	local	sName = UIDropDownMenu_GetSelectedName(LFGFrameNameDropDown1);
	if (sName and (sName ~= "")) then
		bGotLFG = true;
	else
		if ((sEvent == "PARTY_LEFT_RESTORE") and not SaneLFGGlobal.Cfg.KeepLockedOutEntries) then
			local	oSaved = SaneLFGDropdown_BuildSaved();
			if (next(oSaved) ~= nil) then
				local	bDrop, i = false;
				for i = 1, 3 do
					local	oData = SaneLFGPerChar.LFG[i];
					if (oData and oData.Type and ((oData.Type == 3) or (oData.Type == 6)) and oData.Name and (oData.Name ~= 0) and oData.DescLoc and (oData.DescLoc ~= "")) then
						if (SaneLFGObj.bDebug) then
							CHATFRAME:AddMessage("SaneLFG/DEBUG: Lockout for [" .. i .. "] " .. oData.DescLoc .. "?");
						end
						if (SaneLFGDropdown_CheckNameInSaved(oData.Type, strlower(oData.DescLoc), oSaved)) then
							CHATFRAME:AddMessage("SaneLFG: |cFFFF4040Dropping <" .. oData.DescLoc .. ">|r as you seem to be saved to it. If you're *not* done with the instance, you'll have to re-add it to LFG yourself!");
							bDrop = true;
							SaneLFGPerChar.LFG[i] = nil;
						end
					end
				end

				if (bDrop) then
					local	oLFGNew, iCnt = {}, 0;
					for i = 1, 3 do
						if (SaneLFGPerChar.LFG[i]) then
							iCnt = iCnt + 1;
							oLFGNew[iCnt] = SaneLFGPerChar.LFG[i];
						end
					end
					SaneLFGPerChar.LFG = oLFGNew;
				end
			end
		end

		local	t, n = {}, {};
		t[1], n[1], t[2], n[2], t[3], n[3], t[4], n[4], sComment = GetLookingForGroup();

		local	bChanged, bHaveData, i = false, false;
		for i = 1, 3 do
			local	oData = SaneLFGPerChar.LFG[i];
			if (oData and oData.Name and (oData.Name ~= 0)) then
				bHaveData = true;

				SaneLFGObj.MenuEntrySet(true, i, oData.Type, oData.Name);
				SetLookingForGroup(i, oData.Type, oData.Name, LFGFrameLeaderCheckButton:GetChecked(), LFGFrameRoleButton2:GetChecked(), LFGFrameRoleButton3:GetChecked(), LFGFrameRoleButton1:GetChecked());
				if ((t[i] ~= oData.Type) or (n[i] ~= oData.Name)) then
					bChanged = true;
				end
			else
				break
			end
		end

		t[1], n[1], t[2], n[2], t[3], n[3] = GetLookingForGroup();
		bGotLFG = n[1] + n[2] + n[3] > 0;
		if (bGotLFG) then
			LFGFrame_Update();
			if (bChanged) then
				CHATFRAME:AddMessage("SaneLFG: Tried to restore previous LFG choices... (" .. t[1] .. "/" .. n[1] .. ", " .. t[2] .. "/" .. n[2] .. ", " .. t[3] .. "/" .. n[3] .. ")");
			else
				CHATFRAME:AddMessage("SaneLFG: LFG already setup. (" .. t[1] .. "/" .. n[1] .. ", " .. t[2] .. "/" .. n[2] .. ", " .. t[3] .. "/" .. n[3] .. ")");
			end
		elseif (bHaveData) then
			if (bUpdate or bInParty) then
				CHATFRAME:AddMessage("SaneLFG: Failed to restore previous LFG choices - you're in a group.");
			else
				CHATFRAME:AddMessage("SaneLFG: Failed to restore previous LFG choices. :-(");
			end
		end
	end

	if (bGotLFG and SaneLFGPerChar.Comment and (GetNumPartyMembers() == 0) and (GetNumRaidMembers() == 0)) then
		if (sComment and (sComment == SaneLFGPerChar.Comment)) then
			CHATFRAME:AddMessage("SaneLFG: LFG comment already setup.");
		else
			CHATFRAME:AddMessage("SaneLFG: Tried to restore previous LFG comment.");
			SetLFGComment(SaneLFGPerChar.Comment);
		end
	end

	-- LFM parts

	if (SaneLFGPerChar.LFM and (bGotLFG or (GetNumPartyMembers() > 0))) then
		if (bGotLFG) then
--			LFGParentFrameTab2_OnClick();
		else
			CHATFRAME:AddMessage("SaneLFG: Restoring while being in group probably not functional yet. :-(");
		end

		local	selectedLFMType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
		local	selectedLFMName = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
		local	storedLFMName = SaneLFGPerChar.LFM.Name or 0;
		if (storedLFMName ~= 0) then
			if ((selectedLFMType ~= SaneLFGPerChar.LFM.Type) or (selectedLFMName ~= storedLFMName)) then
				SaneLFGObj.MenuEntrySet(false, nil, SaneLFGPerChar.LFM.Type, SaneLFGPerChar.LFM.Name);

				-- hardware event if un-pausing (button clicked or command via keyboard)
				local	sRestored = "(Tried to) restore";
				local	sReOpen = " - You'll have to (re-)open the window to complete LFM initialization.";
				if (sEvent == "CommandHandler") then
					if (CanSendLFGQuery(selectedLFMType, selectedLFMName)) then
						sRestored = "Restored";
						sReOpen = "";
						LFGQuery(selectedLFMType, selectedLFMName);
					end
				end

				LFMFrame_Update();

				local	sName = SaneLFGObj.TypeNameToString(SaneLFGPerChar.LFM.Type, SaneLFGPerChar.LFM.Name);
				CHATFRAME:AddMessage("SaneLFG: " .. sRestored .. " previous LFM choice... (" .. SaneLFGPerChar.LFM.Type .. "/" .. SaneLFGPerChar.LFM.Name .. " <" .. sName .. ">)" .. sReOpen);
			else
				CHATFRAME:AddMessage("SaneLFG: LFM already setup... (" .. SaneLFGPerChar.LFM.Type .. "/" .. SaneLFGPerChar.LFM.Name .. ")");
			end
		end
	end

	iRestoreFromParty = nil;

	iState = iStateNew;
	bStore = bStoreNew;
end

--
--
--

function	SaneLFGObj.Status()
	local	sMsg = "iState = " .. tostring(iState) .. ", bStore = " .. tostring(bStore) .. ", bDing = " .. tostring(bDing) .. ", bDebug = " .. tostring(SaneLFGObj.bDebug) .. ", bPause = " .. tostring(bPause) .. ", iPartyChanged = " .. tostring(iPartyChanged) .. ", iRestoreFromParty = " .. tostring(iRestoreFromParty);
	CHATFRAME:AddMessage("SaneLFG: Status " .. sMsg);
end

--
--
--

function	SaneLFGObj.Debug()
	local iCnt, i, X, s, k, v = GetNumSavedInstances();
	for i = 1, iCnt do
		X = { GetSavedInstanceInfo(i) };

		s = "";
		for k, v in pairs(X) do
			s = s .. ", <" .. tostring(v) .. ">";
		end

		CHATFRAME:AddMessage("SaneLFG/DEBUG: [" .. i .. "] " .. (strsub(s, 3) or ""));
	end
end

function	SaneLFGObj.CheckLock(sName, iType)
	local	oSaved = SaneLFGDropdown_BuildSaved();
	local	s, k, v = "";
	for k, v in pairs(oSaved) do
		s = s .. ", <" .. k .. ": ...>";
	end

	local	bSaved = SaneLFGDropdown_CheckNameInSaved(iType, sName, oSaved);
	CHATFRAME:AddMessage("SaneLFG/DEBUG: " .. iType .. ":" .. sName .. " in " .. (strsub(s, 3) or "<>") .. "? " .. tostring(bSaved));
end

