
-- stuff for multi-mixed support

local	CHATFRAME = ChatFrame2;

--
--
--

-- this holds *everything* we see
local SaneLFMCache = {};

--
--
--

-- overwrite!
local	LFMButton_OnClickOriginal = LFMButton_OnClick;
function LFMButton_OnClick(self, button)
	if (self.isTitle ~= true) then
		LFMButton_OnClickOriginal(self, button);
	end
end

-- overwrite!
local	SaneLFMButton_OnEnter;
local	LFMButton_OnEnterOriginal = LFMButton_OnEnter;
function LFMButton_OnEnter(self)
	if (not SaneLFGGlobal.Cfg.CustomTip and (self.isTitle == nil) and (not self.partyMembers or (self.partyMembers <= 0))) then
		LFMButton_OnEnterOriginal(self);
	else
		-- party members are drawn *in* the OnEnter, not preset while caching :(
		SaneLFMButton_OnEnter(self);
	end
end

local	function	SaneClassToColoredClass(classLocalized)
	local	class = classLocalized;
	if (SaneLFGGlobal.ClassXXToEN) then
		local	classEN = SaneLFGGlobal.ClassXXToEN[classLocalized];
		if (classEN) then
			local	oColor = RAID_CLASS_COLORS[classEN];
			if (oColor) then
				class = format("\124cff%.2x%.2x%.2x%s\124r", oColor.r * 255, oColor.g * 255, oColor.b * 255, classLocalized);
			end
		end
	end

	return class;
end

function SaneLFMButton_OnEnter(self)
	GameTooltip:SetOwner(self, "ANCHOR_RIGHT", 27, -37);
	if (self.isTitle) then
		GameTooltip:SetText("LFG/LFM for " .. self.nameLine, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
		GameTooltip:AddLine(self.criteria);
		GameTooltip:Show();
		return
	end

	if ( self.isLFM ) then
		GameTooltip:SetText(LFM_TITLE, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
	else
		GameTooltip:SetText(LFG_TITLE, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
	end

	local	bCached, oData = false;
	if (self.oIndicesGroup and self.iEntry) then
		oData = SaneLFMCache[self.oIndicesGroup.iType][self.oIndicesGroup.iChoice][self.iEntry];
		if (oData) then
			bCached = true;
		end
	end

	-- class colorization for leader
	local	nameLine = self.nameLine;
	if (not bCached) then
		local	name, level, _, class = GetLFGResultsProxy(self.lfgIndex);
		nameLine = format(LFM_NAME_TEMPLATE, name, level, SaneClassToColoredClass(class));
	end
	GameTooltip:AddLine(nameLine);

	local numPartyMembers = self.partyMembers;
	if ( numPartyMembers > 0 ) then
		GameTooltip:AddTexture("Interface\\GroupFrame\\UI-Group-LeaderIcon");
		-- Only show party members if there are 10 or less
		if ( numPartyMembers > 9 ) then
			GameTooltip:AddLine(format(LFM_NUM_RAID_MEMBER_TEMPLATE, numPartyMembers));
			-- Bogus texture to make the spacing correct
			GameTooltip:AddTexture("");
		else
			local name, level, class;
			for i = 1, numPartyMembers do
				if (not bCached) then
					name, level, class = GetLFGPartyResultsProxy(self.lfgIndex, i);
				else
					name = oData.partyMemberInfo[i][1];
				end

				if ( name ) then
					if (bCached) then
						level = oData.partyMemberInfo[i][2];
					end
					if ( level == "" ) then
						level = "??";
					end

					if (bCached) then
						class = oData.partyMemberInfo[i][3];
					end

					GameTooltip:AddLine(format(LFM_NAME_TEMPLATE, name, level, SaneClassToColoredClass(class)));

					-- Bogus texture to make the spacing correct
					GameTooltip:AddTexture("");
				end
			end
		end
	end

	GameTooltip:AddLine("\n" .. self.criteria, HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);

	if ( self.comment and self.comment ~= "" ) then
		local	oColor = HIGHLIGHT_FONT_COLOR;
		if (SaneLFGGlobal.Cfg.CommentColor) then
			oColor = SaneLFGGlobal.Cfg.CommentColor;
		end
		GameTooltip:AddLine("\n"..self.comment, oColor.r, oColor.g, oColor.b, 1);
	end

	if ( numPartyMembers <= 0 ) then
		GameTooltip:AddLine("\n"..LFG_TOOLTIP_ROLES)
		if ( self.willBeDPS ) then
			GameTooltip:AddLine(DAMAGER);
			GameTooltip:AddTexture("Interface\\LFGFrame\\LFGRole", 0.25, 0.5, 0, 1);
		end
		if ( self.willBeTank ) then
			GameTooltip:AddLine(TANK);
			GameTooltip:AddTexture("Interface\\LFGFrame\\LFGRole", 0.5, 0.75, 0, 1);
		end
		if ( self.willBeHealer ) then
			GameTooltip:AddLine(HEALER);
			GameTooltip:AddTexture("Interface\\LFGFrame\\LFGRole", 0.75, 1, 0, 1);
		end
		if ( self.willBeLeader ) then
			GameTooltip:AddLine(LEADER);
			GameTooltip:AddTexture("Interface\\LFGFrame\\LFGRole", 0, 0.25, 0, 1);
		end
	end
	
	GameTooltip:Show();
end

--
--
--

local	iLFMUpdateButtonElapsed = 0;

local	oLFMUpdatebutton = CreateFrame("Button", "SaneLFMUpdate", LFMFrame, "UIPanelButtonTemplate");

local	function	SaneLFMUpdate_FindOldest(oTypes, oNames, iMax, bInitMissing)
	local	bMissing, iOldestTime, iOldestIx, i = false, time() + 1;
	for i = 1, iMax do
		if (oNames[i] and (oNames[i] ~= 0)) then
			if ((SaneLFMCache[oTypes[i]] == nil) or (SaneLFMCache[oTypes[i]][oNames[i]] == nil)) then
				bMissing = true;
				if (bInitMissing) then
					if (CanSendLFGQuery(oTypes[i], oNames[i])) then
						bInitMissing = false;	-- only one call per click
						LFGQuery(oTypes[i], oNames[i]);
						CHATFRAME:AddMessage("SaneLFG: (Still) Initializing choice " .. oTypes[i] .. "/" .. oNames[i] .. "...");
					elseif (SaneLFGObj.bDebug) then
						CHATFRAME:AddMessage("SaneLFG/DEBUG: Init not allowed (yet)...");
					end
				end
			else
				local	oData = SaneLFMCache[oTypes[i]][oNames[i]];
				if (oData.UpdateTimeAbs < iOldestTime) then
					iOldestIx = i;
					iOldestTime = oData.UpdateTimeAbs;
				end
			end
		end
	end

	if (bMissing and bInitMissing) then
		CHATFRAME:AddMessage("SaneLFG: Failed to initalize any valid choice at all. :-(");
	end

	return bMissing, iOldestIx, iOldestTime;
end

local	function	SaneLFMUpdate_OnClickLeft(oFrame, sMouseBtn)
	local	t, n = {}, {};
	if (SaneLFGPerChar.LFMCycleSet ~= 4) then
		t[1], n[1], t[2], n[2], t[3], n[3] = GetLookingForGroup();
	else
		local	iCnt = 0;
		for i = 1, 3 do
			oData = SaneLFGPerChar.LFG[i];
			if (oData and oData.Name and (oData.Name ~= 0)) then
				iCnt = iCnt + 1;
				t[iCnt] = oData.Type;
				n[iCnt] = oData.Name;
			end
		end
		while (iCnt < 3) do
			iCnt = iCnt + 1;
			t[iCnt] = 1;
			n[iCnt] = 0;
		end
	end

	local	bHaveCategory = 1 ~= UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
	local	bWalk = bHaveCategory and SaneLFGPerChar.LFMCycleSet and (SaneLFGPerChar.LFMCycleSet ~= 2) and (SaneLFGPerChar.LFMCycleSet ~= 4);
	if (n[1] + n[2] + n[3] == 0) then
		bWalk = true;
	end
	if (bWalk) then
		local selectedLFMType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
		local selectedLFMName = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
		if (selectedLFMType and (selectedLFMType > 1)) then
			local	bFound = false;
			if (SaneLFGPerChar.LFMCycleSet and (SaneLFGPerChar.LFMCycleSet == 3) and SaneLFGPerChar.LFMSet and (#SaneLFGPerChar.LFMSet > 0)) then
				local	iCnt, iNext, i = #SaneLFGPerChar.LFMSet, 1;
				for i = 1, iCnt do
					local	oSet = SaneLFGPerChar.LFMSet[i];
					if ((oSet.Type == selectedLFMType) and (oSet.Name == selectedLFMName)) then
						if (IsShiftKeyDown()) then
							iNext = i - 1;
							if (iNext < 1) then
								iNext = iCnt;
							end
						else
							iNext = i + 1;
							if (iNext > iCnt) then
								iNext = 1;
							end
						end
						break;
					end
				end

				if (iNext) then
					if (SaneLFGObj.bDebug) then
						CHATFRAME:AddMessage("SaneLFG/DEBUG: Cycling LFG set choice...");
					end

					bFound = true;
					local	oSet = SaneLFGPerChar.LFMSet[iNext];
					if (oSet) then
						selectedLFMType = oSet.Type;
						selectedLFMName = oSet.Name;
					end
				end
			end

			if (not bFound) then
				if (SaneLFGObj.bDebug) then
					CHATFRAME:AddMessage("SaneLFG/DEBUG: Cycling category choice as a whole...");
				end

				local	iMax = SaneLFGObj.TypeNameCount(selectedLFMType);
				if (IsShiftKeyDown()) then
					selectedLFMName = (selectedLFMName or 0) - 1;
					if (selectedLFMName < 1) then
						selectedLFMName = iMax;
					end
				else
					selectedLFMName = (selectedLFMName or 0) + 1;
					if (selectedLFMName > iMax) then
						selectedLFMName = 1;
					end
				end
			end

			if (CanSendLFGQuery(selectedLFMType, selectedLFMName)) then
				SaneLFGObj.MenuEntrySet(false, nil, selectedLFMType, selectedLFMName);
				LFGQuery(selectedLFMType, selectedLFMName);
			end
		else
			CHATFRAME:AddMessage("SaneLFG: You need to select a valid category first to cycle through...");
		end

		return
	end

	local	iMax, bCustom = 3, false;
	if (not bHaveCategory and SaneLFGPerChar.Cfg.MixedFromCustomList and SaneLFGPerChar.LFMSet) then
		local	iMaxCustom = #SaneLFGPerChar.LFMSet;
		if (iMaxCustom > 0) then
			t = {};
			n = {};
			local	i;
			for i = 1, iMaxCustom do
				local	oSet = SaneLFGPerChar.LFMSet[i];
				t[i] = oSet.Type;
				n[i] = oSet.Name;
			end

			bCustom = true;
			iMax = iMaxCustom;
			if (SaneLFGObj.bDebug) then
				CHATFRAME:AddMessage("SaneLFG/DEBUG: Cycling custom LFM choices...");
			end
		end
	end

	if (not bCustom and SaneLFGObj.bDebug) then
		CHATFRAME:AddMessage("SaneLFG/DEBUG: Cycling LFG choices...");
	end

	local	iNumSet = 0;
	for i = 1, iMax do
		if (n[i] and (n[i] ~= 0)) then
			iNumSet = iNumSet + 1;
		end
	end

	if (iNumSet == 0) then
		CHATFRAME:AddMessage("SaneLFG: LFG choice is none...");
		return
	end

	local	bMissing, iOldestIx, iOldestTime = SaneLFMUpdate_FindOldest(t, n, iMax, true);
	if (bMissing) then
		return
	end

	local	iType, iChoice;
	if (iOldestIx) then
		iType = t[iOldestIx];
		iChoice = n[iOldestIx];
	end

	if (iType and iChoice) then
		if (CanSendLFGQuery(iType, iChoice)) then
			if (bHaveCategory) then
				SaneLFGObj.MenuEntrySet(false, nil, iType, iChoice);
			else
				SetLookingForMore(iType, iChoice);
			end

			local	sName = SaneLFMCache[iType][iChoice].ZoneName;
			CHATFRAME:AddMessage("SaneLFG: Requesting update for " .. iType .. "/" .. iChoice .. " <" .. sName .. ">.");
			LFGQuery(iType, iChoice);
		end
	else
		CHATFRAME:AddMessage("SaneLFG: Hm. Nothing to update?!");
	end
end

local	oLFGMUpdatebuttonMenu = CreateFrame("Frame", "SaneLFMUpdatebuttonMenu", nil, "UIDropDownMenuTemplate");

local	function	SaneLFGColorpickerDone()
	if (not ColorPickerFrame:IsShown()) then
		ColorPickerFrame.func = nil;
		ColorPickerFrame.cancelFunc = nil;
		if (InterfaceOptionsFrame:GetFrameStrata() ~= "HIGH") then
			ColorPickerFrame:SetFrameStrata("DIALOG");
		end
	end
end

local	function	SaneLFGColorpickerDo()
	local	r, g, b = ColorPickerFrame:GetColorRGB();
	local	oColor = SaneLFGGlobal.Cfg.CommentColor;
	if (oColor) then
		oColor.r = r;
		oColor.g = g;
		oColor.b = b;
	else
		oColor = { r = r, g = g, b = b };
		SaneLFGGlobal.Cfg.CommentColor = oColor;
	end

	SaneLFGColorpickerDone();
end

local	function	oLFGMUpdatebuttonMenu_Execute(oFrame, arg1, arg2)
	if (arg1 == nil) then
		return
	end

	if ((arg1 == 10) or (arg1 == 15) or (arg1 == 20) or (arg1 == 25)) then
		SaneLFGPerChar.LFMCycleSet = arg1 / 5 - 1;
	end

	if ((arg1 == 30) or (arg1 == 40)) then
		-- add/remove
		local selectedLFMType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
		local selectedLFMName = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
		if (selectedLFMType and (selectedLFMType > 1)) then
			local	iFound = nil;
			if (SaneLFGPerChar.LFMSet and (#SaneLFGPerChar.LFMSet > 0)) then
				local	iCnt, i = #SaneLFGPerChar.LFMSet, 1;
				for i = 1, iCnt do
					local	oSet = SaneLFGPerChar.LFMSet[i];
					if ((oSet.Type == selectedLFMType) and (oSet.Name == selectedLFMName)) then
						iFound = i;
						break;
					end
				end
			end

			if (iFound) then
				if (arg1 == 40) then
					tremove(SaneLFGPerChar.LFMSet, iFound);
				end
			else
				if (arg1 == 30) then
					if (SaneLFGPerChar.LFMSet == nil) then
						SaneLFGPerChar.LFMSet = {};
					end
					local	sName = SaneLFGObj.TypeNameToString(selectedLFMType, selectedLFMName);
					tinsert(SaneLFGPerChar.LFMSet, { Type = selectedLFMType, Name = selectedLFMName, DescLoc = sName });
				end
			end
		end
	end

	if (arg1 == 50) then
		SaneLFGPerChar.LFMSet = nil;
	end

	if (arg1 == 60) then
		if (SaneLFGPerChar.LFMSet and (#SaneLFGPerChar.LFMSet >= arg2)) then
			local	oEntry = SaneLFGPerChar.LFMSet[arg2];
			if (CanSendLFGQuery(oEntry.Type, oEntry.Name)) then
				SaneLFGObj.MenuEntrySet(false, nil, oEntry.Type, oEntry.Name);
				LFGQuery(oEntry.Type, oEntry.Name);
			end
		end
	end

	if (arg1 == 100) then
		SaneLFGGlobal.Cfg.PauseOnLeavingGroups = not SaneLFGGlobal.Cfg.PauseOnLeavingGroups;
	end
	if (arg1 == 110) then
		SaneLFGGlobal.Cfg.PauseSave = not SaneLFGGlobal.Cfg.PauseSave;
	end

	if (arg1 == 115) then
		SaneLFGGlobal.Cfg.KeepLockedOutEntries = not SaneLFGGlobal.Cfg.KeepLockedOutEntries;
	end

	if (arg1 == 120) then
		SaneLFGGlobal.Cfg.CustomTip = not SaneLFGGlobal.Cfg.CustomTip;
	end
	if (arg1 == 130) then
		if (not ColorPickerFrame:IsShown()) then
			local	oColor = SaneLFGGlobal.Cfg.CommentColor;
			if (oColor == nil) then
				oColor = { r = 0.7, g = 0.95, b = 1 };
			end

			ColorPickerFrame.func = nil;
			ColorPickerFrame:SetColorRGB(oColor.r, oColor.g, oColor.b);
			if (InterfaceOptionsFrame:GetFrameStrata() ~= "HIGH") then
				ColorPickerFrame:SetFrameStrata("FULLSCREEN_DIALOG");
			end

			ColorPickerFrame.func = SaneLFGColorpickerDo;			-- KarmaAvEnK.Colorspace.ModifyDo;
			ColorPickerFrame.cancelFunc = SaneLFGColorpickerDone;		-- KarmaAvEnK.Colorspace.ModifyDone;
			ColorPickerFrame:Show();
		end
	end

	if (arg1 == 200) then
		SaneLFGPerChar.Cfg.PauseOnLogin = not SaneLFGPerChar.Cfg.PauseOnLogin;
	end
	if (arg1 == 210) then
		SaneLFGPerChar.Cfg.MixedFromCustomList = not SaneLFGPerChar.Cfg.MixedFromCustomList;
	end
end

local	function	oLFGMUpdatebuttonMenu_Initializer(oFrame, iLevel)
	iLevel = iLevel or 1;

	local	iLevelIn = iLevel;
	if (oLFGMUpdatebuttonMenu.ClickedMousebutton == "MiddleButton") then
		iLevelIn = iLevel + 1;
	end

	local	info;

	if (iLevelIn == 1) then
		info = {};
		info.text = "SaneLFG: choices menu";
		info.isTitle = 1;
		info.notCheckable = true;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Cycle through complete category";
		info.checked = SaneLFGPerChar.LFMCycleSet == 1;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 10;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Cycle through current LFG choices";
		info.checked = (SaneLFGPerChar.LFMCycleSet == nil) or (SaneLFGPerChar.LFMCycleSet == 2);
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 15;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Cycle through preset list";
		info.checked = SaneLFGPerChar.LFMCycleSet == 3;
		info.disabled = (SaneLFGPerChar.LFMSet == nil) or (#SaneLFGPerChar.LFMSet == 0);
		info.hasArrow = true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 20;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Cycle through stored LFG choices";
		info.checked = SaneLFGPerChar.LFMCycleSet == 4;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 25;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Add current selection to set";
		info.notCheckable = true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 30;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Remove current selection from set";
		info.notCheckable = true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 40;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Clear set";
		info.notCheckable = true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 50;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = " - - - - - - global options - - - - - - ";
		info.justifyH = "CENTER";
		info.notCheckable = true;
		info.disabled = true;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Enter pause mode when leaving a group";
		info.checked = SaneLFGGlobal.Cfg.PauseOnLeavingGroups == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 100;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Save pause status across sessions";
		info.checked = SaneLFGGlobal.Cfg.PauseSave == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 110;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Keep LFG entries, even if saved to it";
		info.checked = SaneLFGGlobal.Cfg.KeepLockedOutEntries == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 115;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Customize tooltip (slightly)";
		info.checked = SaneLFGGlobal.Cfg.CustomTip == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 120;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Modify customized comment color";
		info.notCheckable = true;
		info.disabled = SaneLFGGlobal.Cfg.CustomTip ~= true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 130;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = " - - - - - - per char options - - - - - - ";
		info.justifyH = "CENTER";
		info.notCheckable = true;
		info.disabled = true;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "On login, start with PAUSEd LFG";
		info.checked = SaneLFGPerChar.Cfg.PauseOnLogin == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 200;

		UIDropDownMenu_AddButton(info, iLevel);

		info = {};
		info.text = "Mixed mode: use custom LFM set (not current LFG set)";
		info.checked = SaneLFGPerChar.Cfg.MixedFromCustomList == true;
		info.func = oLFGMUpdatebuttonMenu_Execute;
		info.arg1 = 210;

		UIDropDownMenu_AddButton(info, iLevel);
	end

	if (iLevelIn == 2) then
		if (SaneLFGPerChar.LFMSet and (#SaneLFGPerChar.LFMSet > 0)) then
			local selectedLFMType = UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
			local selectedLFMName = UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);

			local	iCnt, i = #SaneLFGPerChar.LFMSet;
			for i = 1, iCnt do
				info = {};
				local	oEntry = SaneLFGPerChar.LFMSet[i];
				local	sType = SaneLFGObj.TypeToString(oEntry.Type);
				local	sName = SaneLFGObj.TypeNameToString(oEntry.Type, oEntry.Name);
				info.text = "[" .. i .. "] " .. sType .. ": " .. sName;
				info.checked = (oEntry.Type == selectedLFMType) and (oEntry.Name == selectedLFMName);
				info.func = oLFGMUpdatebuttonMenu_Execute;
				info.arg1 = 60;
				info.arg2 = i;

				UIDropDownMenu_AddButton(info, iLevel);
			end
		end
	end
end

local	function	oLFGMUpdatebuttonMenu_OnShow(oFrame)
	if (SaneLFGObj.bDebug and oFrame) then
		CHATFRAME:AddMessage("SaneLFG/DEBUG: Showing menu <" .. oFrame:GetName() .. ">...");
	end
	UIDropDownMenu_Initialize(oLFGMUpdatebuttonMenu, oLFGMUpdatebuttonMenu_Initializer, "MENU");
end

oLFGMUpdatebuttonMenu:SetScript("OnShow", oLFGMUpdatebuttonMenu_OnShow);

local	function	SaneLFMUpdate_OnClickRight(oFrame, sMouseBtn)
	if (oLFGMUpdatebuttonMenu.initialize == nil) then
	--	oLFGMUpdatebuttonMenu:SetPoint("TOPLEFT", oFrame, "TOPRIGHT");
		oLFGMUpdatebuttonMenu_OnShow(nil);
	end

	oLFGMUpdatebuttonMenu.ClickedMousebutton = sMouseBtn;
	ToggleDropDownMenu(1, nil, oLFGMUpdatebuttonMenu, oFrame, 0, 0);
end

local	SaneLFMUpdate_OnLeave;

local	function	SaneLFMUpdate_OnClick(oFrame, sMouseBtn)
	SaneLFMUpdate_OnLeave(oFrame);
	if (sMouseBtn == "LeftButton") then
		SaneLFMUpdate_OnClickLeft(oFrame, sMouseBtn);
	elseif ((sMouseBtn == "MiddleButton") or (sMouseBtn == "RightButton")) then
		SaneLFMUpdate_OnClickRight(oFrame, sMouseBtn);
	else
		CHATFRAME:AddMessage("SaneLFG: How did you do that...?");
	end
end

oLFMUpdatebutton:RegisterForClicks("LeftButtonUp", "MiddleButtonDown", "RightButtonDown");
oLFMUpdatebutton:SetScript("OnClick", SaneLFMUpdate_OnClick);

local	bFirstTime = true;
local	bLFMTipShown = false;
local	iLFMUpdateTipEntered = nil;

local	function	SaneLFMUpdate_OnEnter(oFrame)
	if (bFirstTime) then
		-- first time show immediately
		bFirstTime = false;
		iLFMUpdateTipEntered = 10;
	else
		iLFMUpdateTipEntered = 0;
	end
end

-- local
function	SaneLFMUpdate_OnLeave(oFrame)
	if (bLFMTipShown) then
		GameTooltip_Hide();
	end

	iLFMUpdateTipEntered = nil;
	bLFMTipShown = false;
end

oLFMUpdatebutton:SetScript("OnEnter", SaneLFMUpdate_OnEnter);
oLFMUpdatebutton:SetScript("OnLeave", SaneLFMUpdate_OnLeave);

local	bUpdateDisabled = false;
local	iLFMUpdateCacheElapsed = 0;

local	function	SaneLFMUpdate_OnUpdate(oFrame, iElapsed)
	if (bUpdateDisabled) then
		iLFMUpdateButtonElapsed = iLFMUpdateButtonElapsed + iElapsed;
		if (iLFMUpdateButtonElapsed > 3) then
			bUpdateDisabled = false;
			oLFMUpdatebutton:Enable();
		end
	end

	iLFMUpdateCacheElapsed = iLFMUpdateCacheElapsed + iElapsed;
	if (iLFMUpdateCacheElapsed > LFM_LIST_REFRESH_UPDATE_TIME) then
		iLFMUpdateCacheElapsed = 0;
		SaneLFGObj.LFMFrame_FullCacheAndUpdate();
	end

	if (iLFMUpdateTipEntered) then
		iLFMUpdateTipEntered = iLFMUpdateTipEntered + iElapsed;
		if (iLFMUpdateTipEntered > 5) then
			iLFMUpdateTipEntered = nil;
			bLFMTipShown = true;
			GameTooltip:SetOwner(oFrame, "ANCHOR_BOTTOMRIGHT");
			GameTooltip:AddLine("SaneLFG: cycle through various LFM lists");
			GameTooltip:AddLine(" ");
			GameTooltip:AddLine("Left click to cycle through current choice");
			GameTooltip:AddLine("Shift-left click to cycle backwards");
			GameTooltip:AddLine("Middle click to select from preset list");
			GameTooltip:AddLine("Right click to modify current choice");
			GameTooltip:AddLine(" ");
			GameTooltip:AddLine("Default choices depend on selected category:");
			local	sTypeNone = SaneLFGObj.TypeToString(1);
			GameTooltip:AddLine("'" .. sTypeNone .. "' (lfg mixed mode) -> your lfg entries");
			GameTooltip:AddLine("otherwise -> preset list or all of the category");

			if (#SaneLFGPerChar.LFG > 0) then
				GameTooltip:AddLine(" ");
				GameTooltip:AddLine("Saved LFG selection:");
				local	iCnt, i = #SaneLFGPerChar.LFG;
				for i = 1, iCnt do
					local	oLFG = SaneLFGPerChar.LFG[i];
					local	sType = SaneLFGObj.TypeToString(oLFG.Type);
					local	sName = SaneLFGObj.TypeNameToString(oLFG.Type, oLFG.Name);
					GameTooltip:AddLine("[" .. i .. "] " .. sType .. ": " .. sName);
				end
			end

			GameTooltip:AddLine(" ");
			GameTooltip:AddLine("Current LFG/LFM status:");

			local	oResult, bLFGorLFM = SaneLFGObj.LFGStatusHelper();
			if (bLFGorLFM) then
				local	iCnt, i = #oResult;
				for i = 1, iCnt do
					GameTooltip:AddLine(oResult[i]);
				end
			else
				GameTooltip:AddLine(" Neither LFG nor LFM currently.");
			end

			GameTooltip:Show();
		end
	end
end

oLFMUpdatebutton:SetScript("OnUpdate", SaneLFMUpdate_OnUpdate);

oLFMUpdatebutton:SetText("Cycle");
oLFMUpdatebutton:SetPoint("TOPRIGHT", LFMFrame, "TOPRIGHT", -72, -33);		-- -33, -33 = TOPRIGHT, but got the eye there as well
oLFMUpdatebutton:SetWidth(75);
oLFMUpdatebutton:SetHeight(25);
oLFMUpdatebutton:Show();

--

local	iClicked = 0;

function	SaneLFGObj.BindingCycle()
	if (oLFMUpdatebutton:IsVisible()) then
		if (not bUpdateDisabled) then
			SaneLFMUpdate_OnClickLeft(oLFMUpdatebutton, "LeftButton");
		end
	else
		-- a hidden button does not run its OnUpdate, so we'd stay locked
		-- check negative marker to find if it was kept hidden, then use a direct time check
		local	iTime = GetTime();
		if (not bUpdateDisabled or (iLFMUpdateButtonElapsed == - 0.1) and (iTime - iClicked > 5)) then
			iClicked = GetTime();
			SaneLFMUpdate_OnClickLeft(oLFMUpdatebutton, "LeftButton");
			iLFMUpdateButtonElapsed = - 0.1;
		end
	end
end

--
--
--

local	iLFMType, iLFMChoice;
local	iUpdate = 0;

local	function	SaneLFGQuery(iType, iChoice)
	-- we can't know if this is actually sent, so this *might* be wrong :(
	-- still, it's our best guess for now (GetLookingForGroup() yiels garbage)
	iLFMType, iLFMChoice = iType, iChoice;
	iUpdate = GetTime();

	iLFMUpdateButtonElapsed = 0;
	bUpdateDisabled = true;
	oLFMUpdatebutton:Disable();
end

hooksecurefunc("LFGQuery", SaneLFGQuery);

--
--
--

function SaneLFGObj.LFMFrame_FullCacheAndUpdate(bEvent)
	if (not iLFMType or not iLFMChoice) then
		return
	end

	local name, level, zone, class, criteria1, criteria2, criteria3, comment, numPartyMembers, isLFM;
	local classFileName, willBeLeader, willBeTank, willBeHealer, willBeDPS;
	local selectedLFMType = iLFMType;	-- UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown);
	local selectedLFMName = iLFMChoice;	-- UIDropDownMenu_GetSelectedID(LFMFrameNameDropDown);
	local numResults, totalCount = GetNumLFGResults(selectedLFMType, selectedLFMName);

	if (SaneLFGObj.bDebug and bEvent) then
		CHATFRAME:AddMessage("SaneLFG/DEBUG: " .. selectedLFMType .. "/" .. selectedLFMName .. " => " .. numResults);
	end

	SaneLFMCache[selectedLFMType] = SaneLFMCache[selectedLFMType] or {};
	SaneLFMCache[selectedLFMType][selectedLFMName] = SaneLFMCache[selectedLFMType][selectedLFMName] or {};
	local	oSaneLFMCache = SaneLFMCache[selectedLFMType][selectedLFMName];

	local currIndex = 0;
	for resultIndex = 1, numResults do
		name, level, zone, class, criteria1, criteria2, criteria3, comment, numPartyMembers, isLFM, classFileName, willBeLeader, willBeTank, willBeHealer, willBeDPS = GetLFGResults(selectedLFMType, selectedLFMName, resultIndex);
		if (name) then
			currIndex = currIndex + 1;
			if ( not oSaneLFMCache[currIndex] ) then
				oSaneLFMCache[currIndex] = {};
			end

			local tempTab = oSaneLFMCache[currIndex];
			tempTab[1] = name;
			tempTab[2] = level;
			tempTab[3] = zone;
			tempTab[4] = class;
			tempTab[5] = criteria1;
			tempTab[6] = criteria2;
			tempTab[7] = criteria3;
			tempTab[8] = comment;
			tempTab[9] = numPartyMembers;
			tempTab[10] = isLFM;
			tempTab[11] = classFileName;		-- english class name
			tempTab[12] = willBeLeader;
			tempTab[13] = willBeTank;
			tempTab[14] = willBeHealer;
			tempTab[15] = willBeDPS;

			-- localized -> english
			if (SaneLFGGlobal.ClassXXToEN == nil) then
				SaneLFGGlobal.ClassXXToEN = {};
			end
			SaneLFGGlobal.ClassXXToEN[class] = classFileName;

			if ( not tempTab.partyMemberInfo ) then
				tempTab.partyMemberInfo = {};
			end

			for i= 1, numPartyMembers do
				name, level, class = GetLFGPartyResults(selectedLFMType, selectedLFMName, resultIndex, i);
				if ( not tempTab.partyMemberInfo[i] ) then
					tempTab.partyMemberInfo[i] = {};
				end

				tempTab.partyMemberInfo[i][1] = name;
				tempTab.partyMemberInfo[i][2] = level;
				tempTab.partyMemberInfo[i][3] = class;
			end
		end
	end

	oSaneLFMCache.NumReturnResults = currIndex;
	if ( selectedLFMName and selectedLFMType ) then
		oSaneLFMCache.ZoneName = select(selectedLFMName * 2 - 1, GetLFGTypeEntries(selectedLFMType));
	else
		oSaneLFMCache.ZoneName = "";
	end

	oSaneLFMCache.UpdateTimeAbs = time();
	oSaneLFMCache.UpdateTimeRel = GetTime();

	if ((1 == UIDropDownMenu_GetSelectedID(LFMFrameTypeDropDown)) and
	    (LFMFrameColumn1Type == LFMCOLUMN1TYPE_ALL)) then
		LFMFrame_Update();
	end
end

function SaneLFGObj.LFMFrame_FullUpdate()
	-- TODO: let user select which of the available data he cares about
	-- currently, show those we're also LFG for

	local	t, n = {}, {};
	t[1], n[1], t[2], n[2], t[3], n[3] = GetLookingForGroup();

	-- build index table of selected data, so we have totals and we can just
	-- walk over the results without further thinking
	local roleFilterTab = LFMFrame_RoleFilter_Individual;

	local	oIndexTable, iIndex = {}, 0;
	local	iType, oType;
	for iType, oType in pairs(SaneLFMCache) do
		local	iChoice, oChoice;
		for iChoice, oChoice in pairs(oType) do
			local	bAdd, i = false;
			if (SaneLFGPerChar.Cfg.MixedFromCustomList) then
				local	iCnt, i = #SaneLFGPerChar.LFMSet;
				for i = 1, iCnt do
					local	oSet = SaneLFGPerChar.LFMSet[i];
					if ((oSet.Type == iType) and (oSet.Name == iChoice)) then
						bAdd = true;
						break;
					end
				end
			else
				for i = 1, 3 do
					if ((t[i] == iType) and (n[i] == iChoice)) then
						bAdd = true;
						break;
					end
				end
			end

			-- only add those that actually have non-zero people
			if (bAdd) then
				local	oIndicesGroup = { iType = iType, iChoice = iChoice };
				iIndex = iIndex + 1;
				oIndexTable[iIndex] = { isTitle = true, oTable = oChoice, oIndicesGroup = oIndicesGroup };
				local	iEntry;
				for iEntry = 1, oChoice.NumReturnResults do
					local	oTable = oChoice[iEntry];
					if ((oTable[9] and (oTable[9] > 0)) or
					    (roleFilterTab["tank"] and oTable[13]) or
					    (roleFilterTab["healer"] and oTable[14]) or
					    (roleFilterTab["damage"] and oTable[15])) then
						iIndex = iIndex + 1;
						oIndexTable[iIndex] = { isTitle = false, oTable = oTable, oIndicesGroup = oIndicesGroup, iEntry = iEntry };
					end
				end
			end
		end
	end

	local numResults = iIndex;
	local scrollOffset = FauxScrollFrame_GetOffset(LFMListScrollFrame);
	if (numResults <= LFGS_TO_DISPLAY) then
		scrollOffset = 0;
	end

	local resultIndex = scrollOffset;
	local showScrollBar = nil;

	if ( numResults > LFGS_TO_DISPLAY ) then
		showScrollBar = 1;
	end

	LFMFrameColumnHeader4Group:Hide();
	LFMFrameColumnHeader4:Show();
	LFMFrameColumnHeader5:Show();
	LFMFrameColumnHeader6:Show();
	LFMFrameColumnHeader7:Show();

	LFMFrameTotals:SetText(format(WHO_FRAME_TOTAL_TEMPLATE, numResults));

	local button, buttonText, buttonLevel, classText, classTextColor;
	for i = 1, LFGS_TO_DISPLAY, 1 do
		resultIndex = scrollOffset + i;
		button = _G["LFMFrameButton"..i];
		button.lfgIndex = resultIndex;
		if ( resultIndex <= numResults ) then
			local	oData = oIndexTable[resultIndex];
			button.isTitle = oData.isTitle;
			button.oIndicesGroup = oData.oIndicesGroup;
			button.iEntry = oData.iEntry;
			if (oData.isTitle) then
				local	sZoneMain, sZoneSub = oData.oTable.ZoneName, "";
				local	iPos = strfind(sZoneMain, " %- ");
				if (iPos) then
					sZoneSub = strsub(sZoneMain, iPos + 3);
					sZoneMain = strsub(sZoneMain, 1, iPos - 1);
				end

				buttonText = _G["LFMFrameButton"..i.."Name"];
				buttonText:SetText("~" .. sZoneMain);
				buttonText:SetTextColor(0, 1, 1);

				local	iAge, xAge = time() - oData.oTable.UpdateTimeAbs;
				if (iAge < 90) then
					xAge = iAge .. "s";
				elseif (iAge < 570) then
					xAge = math.floor((iAge + 30) / 60) .. "m";
				else
					xAge = "oo";
				end

				buttonLevel = _G["LFMFrameButton"..i.."Level"];
				buttonLevel:SetText(xAge);

				classText = _G["LFMFrameButton"..i.."Class"];
				classText:SetText(sZoneSub);
				classText:SetTextColor(0, 1, 1);

				_G["LFMFrameButton"..i.."PartyIcon"]:Hide();

				button.MembersFontString:SetText("++++++++++++");
				button.MembersFontString:Show();
				button.LeaderIcon:Hide();
				button.TankIcon:Hide();
				button.HealerIcon:Hide();
				button.DamageIcon:Hide()

				button.isLFM = nil;
				button.nameLine = oData.oTable.ZoneName;
				button.criteria = "Updated " .. date("%H:%M:%S", oData.oTable.UpdateTimeAbs) .. " (" .. iAge .. " seconds ago)";
				button.comment = nil;
				button.partyMembers = -1;	-- *must* be set (for tooltip)
				button.willBeLeader = nil;
				button.willBeTank = nil;
				button.willBeHealer = nil;
				button.willBeDPS = nil;
				
				LFMFrame.selectedLFM = nil;
				button:UnlockHighlight();
			else
				local	oEntry = oData.oTable;
				if ( oEntry[11] ) then	-- classFileName
					classTextColor = RAID_CLASS_COLORS[oEntry[11]];
				else
					classTextColor = NORMAL_FONT_COLOR;
				end

				buttonText = _G["LFMFrameButton"..i.."Name"];
				buttonText:SetText(oEntry[1]);	-- name
				buttonText:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);

				buttonLevel = _G["LFMFrameButton"..i.."Level"];
				buttonLevel:SetText(oEntry[2]);	-- level

				classText = _G["LFMFrameButton"..i.."Class"];
				classText:SetText(oEntry[4]);	-- class
				classText:SetTextColor(classTextColor.r, classTextColor.g, classTextColor.b);

				-- Show the party leader icon if necessary
				if ( oEntry[9] > 0 ) then	-- numPartyMembers;
					_G["LFMFrameButton"..i.."PartyIcon"]:Show();
				else	
					_G["LFMFrameButton"..i.."PartyIcon"]:Hide();
				end
				
				if ( oEntry[9] > 0 ) then
					button.MembersFontString:SetText(oEntry[9] + 1);
					button.MembersFontString:Show();
					button.LeaderIcon:Hide();
					button.TankIcon:Hide();
					button.HealerIcon:Hide();
					button.DamageIcon:Hide()
				else
					button.MembersFontString:Hide();

					--Show roles as necessary
					if ( oEntry[12] ) then	-- willBeLeader
						button.LeaderIcon:Show();
						if ( LFMFrame_RoleFilter_Individual["leader"] ) then
							button.LeaderIcon:SetTexture("Interface\\LFGFrame\\LFGRole");
						else
							button.LeaderIcon:SetTexture("Interface\\LFGFrame\\LFGRole_BW");
						end
					else
						button.LeaderIcon:Hide();
					end
					if ( oEntry[13] ) then	-- willBeTank
						button.TankIcon:Show();
						if ( LFMFrame_RoleFilter_Individual["tank"] ) then
							button.TankIcon:SetTexture("Interface\\LFGFrame\\LFGRole");
						else
							button.TankIcon:SetTexture("Interface\\LFGFrame\\LFGRole_BW");
						end
					else
						button.TankIcon:Hide();
					end
					if ( oEntry[14] ) then	-- willBeHealer
						button.HealerIcon:Show();
						if ( LFMFrame_RoleFilter_Individual["healer"] ) then
							button.HealerIcon:SetTexture("Interface\\LFGFrame\\LFGRole");
						else
							button.HealerIcon:SetTexture("Interface\\LFGFrame\\LFGRole_BW");
						end
					else
						button.HealerIcon:Hide();
					end
					if ( oEntry[15] ) then	-- willBeDPS
						button.DamageIcon:Show();
						if ( LFMFrame_RoleFilter_Individual["damage"] ) then
							button.DamageIcon:SetTexture("Interface\\LFGFrame\\LFGRole");
						else
							button.DamageIcon:SetTexture("Interface\\LFGFrame\\LFGRole_BW");
						end
					else
						button.DamageIcon:Hide();
					end
				end
				
				-- Set info for the tooltip
				button.isLFM = oEntry[10];	-- isLFM
				button.nameLine = format(LFM_NAME_TEMPLATE, oEntry[1], oEntry[2], SaneClassToColoredClass(oEntry[4]));	-- name, level, class
				button.criteria = BuildNewLineListString(oEntry[5], oEntry[6], oEntry[7]);	-- criteria1, criteria2, criteria3
				if ( not button.criteria ) then
				--	button.criteria = GetLFGCacheSourceInfo();
					button.criteria = "LF?";
				end

				button.comment = oEntry[8];		-- comment
				button.partyMembers = oEntry[9];	-- numPartyMembers
				button.willBeLeader = oEntry[12];	-- willBeLeader
				button.willBeTank = oEntry[13];		-- willBeTank
				button.willBeHealer = oEntry[14];	-- willBeHealer
				button.willBeDPS = oEntry[15];		-- willBeDPS

				-- Highlight the correct lfm
				if ( LFMFrame.selectedName == oEntry[1] ) then	-- name
					LFMFrame.selectedLFM = resultIndex;
					button:LockHighlight();
				else
					button:UnlockHighlight();
				end
			end

			if ( showScrollBar ) then
				classText:SetWidth(70);
			else
				classText:SetWidth(92);
			end

			button:Show();
		else
			button:Hide();
		end
	end

	-- Update send message and group invite buttons
	if ( LFMFrame.selectedName and (LFMFrame.selectedName ~= UnitName("player")) ) then
		LFMFrameSendMessageButton:Enable();
		if ( CanGroupInvite() and LFMFrameColumn1Type ~= LFMCOLUMN1TYPE_GROUP) then
			LFMFrameGroupInviteButton:Enable();
		else
			LFMFrameGroupInviteButton:Disable();
		end
	else
		LFMFrameSendMessageButton:Disable();
		LFMFrameGroupInviteButton:Disable();
	end

	-- If need scrollbar resize columns
	if ( showScrollBar ) then
		WhoFrameColumn_SetWidth(LFMFrameColumnHeader3, 88);
	else
		WhoFrameColumn_SetWidth(LFMFrameColumnHeader3, 110);
	end

	-- ScrollFrame update
	FauxScrollFrame_Update(LFMListScrollFrame, numResults, LFGS_TO_DISPLAY, 16);
end

--
--
--

function	SaneLFGObj.LFGStatusHelper()
	local	oResult = {};
	tinsert(oResult, "LFG: ");

	local	t, n, sComment, isQueued, isLFG, isLFM, i = {}, {};
	t[1], n[1], t[2], n[2], t[3], n[3], t[4], n[4], sComment, isQueued, isLFG, isLFM = GetLookingForGroup();
	for i = 1, 3 do
		local	sType = SaneLFGObj.TypeToString(t[i]);
		local	sName = SaneLFGObj.TypeNameToString(t[i], n[i]);
		tinsert(oResult, " [" .. i .. "] " .. sType .. ": " .. sName);
	end

	tinsert(oResult, "LFM: ");

	i = 4;
	if (n[i] and (n[i] ~= 0)) then
		local	sType = SaneLFGObj.TypeToString(t[i]);
		local	sName = SaneLFGObj.TypeNameToString(t[i], n[i]);
		tinsert(oResult, " " .. sType .. ": " .. sName);
	elseif (iLFMChoice and (iLFMChoice ~= 0)) then
		local	sType = SaneLFGObj.TypeToString(iLFMType);
		local	sName = SaneLFGObj.TypeNameToString(iLFMType, iLFMChoice);
		tinsert(oResult, " (indirect) " .. sType .. ": " .. sName);
	else
		tinsert(oResult, " --");
	end

	tinsert(oResult, "Status: isQueued = " .. tostring(isQueued) .. ", isLFG = " .. tostring(isLFG) .. ", isLFM = " .. tostring(isLFM) .. ".");

	return oResult, isLFG or isLFM;
end

function	SaneLFGObj.LFGStatus()
	local	oResult = SaneLFGObj.LFGStatusHelper();
	local	sMsg, iCnt, i = "", #oResult;
	for i = 1, iCnt do
		sMsg = sMsg .. "\n" .. oResult[i];
	end

	CHATFRAME:AddMessage("SaneLFG: LFG/LFM status =>" .. sMsg);
end

